@@ -1,4 +1,4 @@
-# Copyright (c) 2013-2016 by Ron Frederick <ronf@timeheart.net>.
+# Copyright (c) 2013-2017 by Ron Frederick <ronf@timeheart.net>.
 # All rights reserved.
 #
 # This program and the accompanying materials are made available under
@@ -662,6 +662,40 @@ class SSHKey:
         with open(filename, 'wb') as f:
             f.write(self.export_public_key(*args, **kwargs))
 
+    def append_private_key(self, filename, *args, **kwargs):
+        """Append a private key to a file in the requested format
+
+           This method is a simple wrapper around :meth:`export_private_key`
+           which appends the exported key data to an existing file.
+
+           :param str filename:
+               The filename to append the private key to.
+           :param \\*args,\\ \\*\\*kwargs:
+               Additional arguments to pass through to
+               :meth:`export_private_key`.
+
+        """
+
+        with open(filename, 'ab') as f:
+            f.write(self.export_private_key(*args, **kwargs))
+
+    def append_public_key(self, filename, *args, **kwargs):
+        """Append a public key to a file in the requested format
+
+           This method is a simple wrapper around :meth:`export_public_key`
+           which appends the exported key data to an existing file.
+
+           :param str filename:
+               The filename to append the public key to.
+           :param \\*args,\\ \\*\\*kwargs:
+               Additional arguments to pass through to
+               :meth:`export_public_key`.
+
+        """
+
+        with open(filename, 'ab') as f:
+            f.write(self.export_public_key(*args, **kwargs))
+
 
 class SSHCertificate:
     """Parent class which holds an SSH certificate"""
@@ -970,6 +1004,23 @@ class SSHCertificate:
         with open(filename, 'wb') as f:
             f.write(self.export_certificate(*args, **kwargs))
 
+    def append_certificate(self, filename, *args, **kwargs):
+        """Append a certificate to a file in the requested format
+
+           This function is a simple wrapper around export_certificate
+           which appends the exported certificate to an existing file.
+
+           :param str filename:
+               The filename to append the certificate to.
+           :param \\*args,\\ \\*\\*kwargs:
+               Additional arguments to pass through to
+               :meth:`export_certificate`.
+
+        """
+
+        with open(filename, 'ab') as f:
+            f.write(self.export_certificate(*args, **kwargs))
+
     def validate(self, cert_type, principal):
         """Validate the certificate type, validity period, and principal
 
@@ -1074,31 +1125,6 @@ class SSHKeyPair:
        algorithm and public key or certificate information needed to
        identify what key was used for signing.
 
-       .. attribute:: algorithm
-
-          The public key or certificate algorithm associated with this
-          key pair.
-
-       .. attribute:: sig_algorithm
-
-          The public key signature algorithm currently associated with
-          this key pair.
-
-       .. attribute:: host_key_algorithms
-
-          The list of host key algorithms this key pair is capable of
-          supporting.
-
-       .. attribute:: sig_algorithms
-
-          The list of signature algorithms this key pair is capable of
-          supporting.
-
-       .. attribute:: public_data
-
-          The public key or certificate associated with this key pair
-          in OpenSSH binary format.
-
     """
 
     _key_type = 'unknown'
@@ -2124,6 +2150,8 @@ def load_keypairs(keylist, passphrase=None):
             return [SSHLocalKeyPair(key) for key in keys]
     elif isinstance(keylist, (tuple, bytes, SSHKey, SSHKeyPair)):
         keylist = [keylist]
+    elif not keylist:
+        keylist = []
 
     for key in keylist:
         if isinstance(key, SSHKeyPair):
