@@ -13,6 +13,7 @@
 """SSH channel and session handlers"""
 
 import asyncio
+import binascii
 
 from .constants import DEFAULT_LANG, DISC_PROTOCOL_ERROR, EXTENDED_DATA_STDERR
 from .constants import MSG_CHANNEL_OPEN, MSG_CHANNEL_WINDOW_ADJUST
@@ -20,6 +21,7 @@ from .constants import MSG_CHANNEL_DATA, MSG_CHANNEL_EXTENDED_DATA
 from .constants import MSG_CHANNEL_EOF, MSG_CHANNEL_CLOSE, MSG_CHANNEL_REQUEST
 from .constants import MSG_CHANNEL_SUCCESS, MSG_CHANNEL_FAILURE
 from .constants import OPEN_CONNECT_FAILED, PTY_OP_RESERVED, PTY_OP_END
+from .constants import OPEN_REQUEST_X11_FORWARDING_FAILED
 from .constants import OPEN_REQUEST_PTY_FAILED, OPEN_REQUEST_SESSION_FAILED
 from .editor import SSHLineEditorChannel, SSHLineEditorSession
 from .misc import ChannelOpenError, DisconnectError, map_handler_name
@@ -68,10 +70,12 @@ class SSHChannel(SSHPacketHandler):
         self._recv_buf = []
         self._recv_partial = {}
 
+        self._request_queue = []
+
         self._open_waiter = None
         self._request_waiters = []
 
-        self._close_event = asyncio.Event()
+        self._close_event = asyncio.Event(loop=loop)
 
         self.set_write_buffer_limits()
 
@@ -292,6 +296,36 @@ class SSHChannel(SSHPacketHandler):
         else:
             self._deliver_data(data, datatype)
 
+    def _service_next_request(self):
+        """Process next item on channel request queue"""
+
+        request, packet, _ = self._request_queue[0]
+
+        name = '_process_' + map_handler_name(request) + '_request'
+        handler = getattr(self, name, None)
+        result = handler(packet) if callable(handler) else False
+
+        if result is not None:
+            self._report_response(result)
+
+    def _report_response(self, result):
+        """Report back the response to a previously issued channel request"""
+
+        request, _, want_reply = self._request_queue.pop(0)
+
+        if want_reply and self._send_state not in {'close_pending', 'closed'}:
+            if result:
+                self._send_packet(MSG_CHANNEL_SUCCESS)
+            else:
+                self._send_packet(MSG_CHANNEL_FAILURE)
+
+        if result and request in {'shell', 'exec', 'subsystem'}:
+            self._session.session_started()
+            self.resume_reading()
+
+        if self._request_queue:
+            self._service_next_request()
+
     def process_connection_close(self, exc):
         """Process the SSH connection closing"""
 
@@ -471,19 +505,9 @@ class SSHChannel(SSHPacketHandler):
             raise DisconnectError(DISC_PROTOCOL_ERROR,
                                   'Invalid channel request') from None
 
-        name = '_process_' + map_handler_name(request) + '_request'
-        handler = getattr(self, name, None)
-        result = handler(packet) if callable(handler) else False
-
-        if want_reply and self._send_state not in {'close_pending', 'closed'}:
-            if result:
-                self._send_packet(MSG_CHANNEL_SUCCESS)
-            else:
-                self._send_packet(MSG_CHANNEL_FAILURE)
-
-        if result and request in {'shell', 'exec', 'subsystem'}:
-            self._session.session_started()
-            self.resume_reading()
+        self._request_queue.append((request, packet, want_reply))
+        if len(self._request_queue) == 1:
+            self._service_next_request()
 
     def _process_response(self, pkttype, packet):
         """Process a success or failure response"""
@@ -795,9 +819,18 @@ class SSHClientChannel(SSHChannel):
         self._exit_status = None
         self._exit_signal = None
 
+    def _cleanup(self, exc=None):
+        """Clean up this channel"""
+
+        if self._conn: # pragma: no branch
+            self._conn.detach_x11_listener(self)
+
+        super()._cleanup(exc)
+
     @asyncio.coroutine
-    def create(self, session_factory, command, subsystem, env,
-               term_type, term_size, term_modes, agent_forwarding):
+    def create(self, session_factory, command, subsystem, env, term_type,
+               term_size, term_modes, x11_forwarding, x11_display,
+               x11_auth_path, x11_single_connection, agent_forwarding):
         """Create an SSH client session"""
 
         packet = yield from self._open(b'session')
@@ -843,6 +876,24 @@ class SSHClientChannel(SSHChannel):
                 raise ChannelOpenError(OPEN_REQUEST_PTY_FAILED,
                                        'PTY request failed')
 
+        if x11_forwarding:
+            try:
+                auth_proto, remote_auth, screen = \
+                    yield from self._conn.attach_x11_listener(
+                        self, x11_display, x11_auth_path, x11_single_connection)
+            except ValueError as exc:
+                raise ChannelOpenError(OPEN_REQUEST_X11_FORWARDING_FAILED,
+                                       str(exc)) from None
+
+            result = yield from self._make_request(
+                b'x11-req', Boolean(x11_single_connection), String(auth_proto),
+                String(binascii.b2a_hex(remote_auth)), UInt32(screen))
+
+            if not result:
+                self._conn.detach_x11_listener(self)
+                raise ChannelOpenError(OPEN_REQUEST_X11_FORWARDING_FAILED,
+                                       'X11 forwarding request failed')
+
         if agent_forwarding:
             self._send_request(b'auth-agent-req@openssh.com')
 
@@ -1019,7 +1070,7 @@ class SSHServerChannel(SSHChannel):
     _write_datatypes = {EXTENDED_DATA_STDERR}
 
     def __init__(self, conn, loop, allow_pty, line_editor, line_history,
-                 agent_forwarding, encoding, window, max_pktsize):
+                 encoding, window, max_pktsize):
         """Initialize an SSH server channel"""
 
         super().__init__(conn, loop, encoding, window, max_pktsize)
@@ -1027,20 +1078,27 @@ class SSHServerChannel(SSHChannel):
         self._allow_pty = allow_pty
         self._line_editor = line_editor
         self._line_history = line_history
-        self._agent_forwarding = agent_forwarding
         self._env = self._conn.get_key_option('environment', {})
         self._command = None
         self._subsystem = None
         self._term_type = None
         self._term_size = (0, 0, 0, 0)
         self._term_modes = {}
+        self._x11_display = None
+
+    def _cleanup(self, exc=None):
+        """Clean up this channel"""
+
+        self._conn.detach_x11_listener(self)
+
+        super()._cleanup(exc)
 
     def _wrap_session(self, session):
         """Wrap a line editor around the session if enabled"""
 
         if self._line_editor:
             chan = SSHLineEditorChannel(self, session, self._line_history)
-            session = SSHLineEditorSession(session)
+            session = SSHLineEditorSession(chan, session)
         else:
             chan = self
 
@@ -1095,18 +1153,44 @@ class SSHServerChannel(SSHChannel):
 
         return result
 
+    def _process_x11_req_request(self, packet):
+        """Process request to enable X11 forwarding"""
+
+        _ = packet.get_boolean()                        # single_connection
+        auth_proto = packet.get_string()
+        auth_data = packet.get_string()
+        screen = packet.get_uint32()
+        packet.check_end()
+
+        try:
+            auth_data = binascii.a2b_hex(auth_data)
+        except binascii.Error:
+            return False
+
+        self._conn.create_task(self._finish_x11_req_request(auth_proto,
+                                                            auth_data, screen))
+
+    @asyncio.coroutine
+    def _finish_x11_req_request(self, auth_proto, auth_data, screen):
+        """Finish processing request to enable X11 forwarding"""
+
+        self._x11_display = yield from self._conn.attach_x11_listener(
+            self, auth_proto, auth_data, screen)
+
+        self._report_response(bool(self._x11_display))
+
     def _process_auth_agent_req_at_openssh_dot_com_request(self, packet):
         """Process a request to enable ssh-agent forwarding"""
 
         packet.check_end()
 
-        if not self._agent_forwarding or \
-           not self._conn.check_key_permission('agent-forwarding') or \
-           not self._conn.check_certificate_permission('agent-forwarding'):
-            return False
+        self._conn.create_task(self._finish_agent_req_request())
 
-        self._conn.agent_forwarding_enabled()
-        return True
+    @asyncio.coroutine
+    def _finish_agent_req_request(self):
+        """Finish processing request to enable agent forwarding"""
+
+        self._report_response((yield from self._conn.create_agent_listener()))
 
     def _process_env_request(self, packet):
         """Process a request to set an environment variable"""
@@ -1331,6 +1415,37 @@ class SSHServerChannel(SSHChannel):
 
         return self._term_modes.get(mode)
 
+    def get_x11_display(self):
+        """Return the display to use for X11 forwarding
+
+           When X11 forwarding has been requested by the client, this
+           method returns the X11 display which should be used to open
+           a forwarded connection. If the client did not request X11
+           forwarding, this method returns ``None``.
+
+           :returns: A str containing the X11 display or ``None`` if
+                     X11 fowarding was not requested
+
+        """
+
+        return self._x11_display
+
+    def get_agent_path(self):
+        """Return the path of the ssh-agent listening socket
+
+           When agent forwarding has been requested by the client,
+           this method returns the path of the listening socket which
+           should be used to open a forwarded agent connection. If the
+           client did not request agent forwarding, this method returns
+           ``None``.
+
+           :returns: A str containing the ssh-agent socket path or
+                     ``None`` if agent fowarding was not requested
+
+        """
+
+        return self._conn.get_agent_path()
+
     def set_xon_xoff(self, client_can_do):
         """Set whether the client should enable XON/XOFF flow control
 
@@ -1535,6 +1650,27 @@ class SSHUNIXChannel(SSHForwardChannel):
         self._extra['local_peername'] = dest_path
         self._extra['remote_peername'] = ''
 
+
+class SSHX11Channel(SSHForwardChannel):
+    """SSH X11 channel"""
+
+    @asyncio.coroutine
+    def open(self, session_factory, orig_host, orig_port):
+        """Open an SSH X11 channel"""
+
+        self._extra['local_peername'] = (orig_host, orig_port)
+        self._extra['remote_peername'] = None
+
+        return (yield from self._open(session_factory, b'x11',
+                                      String(orig_host), UInt32(orig_port)))
+
+    def set_inbound_peer_names(self, orig_host, orig_port):
+        """Set local and remote peer name for inbound connections"""
+
+        self._extra['local_peername'] = None
+        self._extra['remote_peername'] = (orig_host, orig_port)
+
+
 class SSHAgentChannel(SSHForwardChannel):
     """SSH agent channel"""
 
